"""Flow engine."""

from .executor import Executor  # noqa: F401
from .flow import Flow, FlowExecError, FlowStatus, get_outer_function  # noqa: F401
from .step import FuncProp, Prop, Step  # noqa: F401



"""User BLoCs."""

from typing import List

from logzero import logger
from sqlalchemy import select

from blackcap.auther import auther_registry
from blackcap.configs import config_registry
from blackcap.db import DBSession
from blackcap.models.protagonist import ProtagonistDB
from blackcap.schemas.api.user.delete import UserDelete
from blackcap.schemas.api.user.get import UserGetQueryParams, UserQueryType
from blackcap.schemas.api.user.post import UserCreate
from blackcap.schemas.api.user.put import UserUpdate
from blackcap.schemas.user import User


config = config_registry.get_config()
auther = auther_registry.get_auther(config.AUTHER)


def create_user(user_create_list: List[UserCreate]) -> List[User]:
    """Create user and save to DB from UserCreate request.

    Args:
        user_create_list (List[UserCreate]): List of UserCreate request

    Raises:
        Exception: error

    Returns:
        List[User]: Created user
    """
    try:
        user_list = auther.register_user(user_create_list)
        return user_list
    except Exception as e:
        logger.error(f"Unable to register users due to {e}")
        raise e


def get_users(query_params: UserGetQueryParams) -> List[User]:
    """Query DB for users.

    Args:
        query_params (UserGetQueryParams): User Query params

    Raises:
        Exception: error

    Returns:
        List[User]: List of Users
    """
    user_list: List[ProtagonistDB] = []

    if query_params.query_type == UserQueryType.GET_ALL_USERS:
        stmt = select(ProtagonistDB)
    if query_params.query_type == UserQueryType.GET_USER_BY_ID:
        stmt = select(ProtagonistDB).where(ProtagonistDB.id == query_params.user_id)
    if query_params.query_type == UserQueryType.GET_USERS_BY_EMAIL:
        stmt = select(ProtagonistDB).where(ProtagonistDB.email == query_params.email)
    if query_params.query_type == UserQueryType.GET_USERS_BY_ORGANISATION:
        stmt = select(ProtagonistDB).where(
            ProtagonistDB.organisation == query_params.organisation
        )

    with DBSession() as session:
        try:
            user_list: List[ProtagonistDB] = session.execute(stmt).scalars().all()
            user_list = [User(user_id=obj.id, **obj.to_dict()) for obj in user_list]
            return user_list
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to fetch users due to {e}")
            raise e


def update_user(user_update: UserUpdate) -> User:
    """Update user in the DB from UserUpdate request.

    Args:
        user_update (UserUpdate): User update request

    Returns:
        User: Instance of updated user
    """
    stmt = select(ProtagonistDB).where(ProtagonistDB.id == user_update.user_id)
    with DBSession() as session:
        try:
            user_list: List[ProtagonistDB] = session.execute(stmt).scalars().all()
            if not user_list:
                # TODO: raise not found error
                pass
            if len(user_list) == 1:
                user_update_dict = user_update.dict(exclude_defaults=True)
                user_update_dict.pop("user_id")
                updated_user = user_list[0].update(session, **user_update_dict)
                return User(user_id=updated_user.id, **updated_user.to_dict())
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to update user: {user_update.dict()} due to {e}")
            # TODO: Raise error


def delete_user(user_delete: UserDelete) -> User:
    """Delete user in the DB from UserDelete request.

    Args:
        user_delete (UserDelete): User delete request

    Raises:
        Exception: error

    Returns:
        User: Instance of deleted user
    """
    stmt = select(ProtagonistDB).where(ProtagonistDB.id == user_delete.user_id)
    with DBSession() as session:
        try:
            user_list: List[ProtagonistDB] = session.execute(stmt).scalars().all()
            if not user_list:
                # TODO: raise not found error
                pass
            if len(user_list) == 1:
                deleted_user = user_list[0].delete(session)
                return User(user_id=deleted_user.id, **deleted_user.to_dict())
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to delete user: {user_delete.dict()} due to {e}")
            raise e



"""Blackcap common schemas."""

from typing import Any, Dict, List

from pydantic import BaseModel


class ResponseSchema(BaseModel):
    """API response schema."""

    msg: str
    items: Dict[str, List[Any]] = {}
    errors: Dict[str, List[Any]] = {}



"""GCP implementation of messenger."""

import json
from typing import Callable, Dict, Optional

from google.auth import jwt
from google.cloud import pubsub_v1
from google.cloud.pubsub_v1.subscriber.message import Message as GCPMessage
from logzero import logger

from blackcap.configs.base import BaseConfig
from blackcap.messenger.base import BaseMessenger
from blackcap.schemas.message import Message
from blackcap.utils.json_encoders import UUIDEncoder


class GCPMessenger(BaseMessenger):
    """GCP(Pub/Sub) implementation of Messenger."""

    CONFIG_KEY_VAL = "GCP"

    def __init__(self: "GCPMessenger", config: BaseConfig) -> None:
        """Initialize Messenger with app config.

        Args:
            config (BaseConfig): Config to initialize messenger
        """
        self.pub_audience = "https://pubsub.googleapis.com/google.pubsub.v1.Publisher"
        self.sub_audience = "https://pubsub.googleapis.com/google.pubsub.v1.Subscriber"

        self.config = config

        self.project_id = config.GCP_PROJECT_ID

    @property
    def service_account_info(self: "GCPMessenger") -> Dict:
        """Service account info."""
        return json.load(open(self.config.GOOGLE_APPLICATION_CREDENTIALS))

    @property
    def publisher(self: "GCPMessenger") -> pubsub_v1.PublisherClient:
        """Publisher Client."""
        self.pub_credentials = jwt.Credentials.from_service_account_info(
            self.service_account_info, audience=self.pub_audience
        )
        return pubsub_v1.PublisherClient(credentials=self.pub_credentials)

    @property
    def subscriber(self: "GCPMessenger") -> pubsub_v1.SubscriberClient:
        """Subscriber Client."""
        self.sub_credentials = jwt.Credentials.from_service_account_info(
            self.service_account_info, audience=self.sub_audience
        )
        return pubsub_v1.SubscriberClient(credentials=self.sub_credentials)

    def publish(self: "GCPMessenger", msg: Dict, topic_id: str) -> str:
        """Publish msg on the GCP Pub/Sub queue.

        Args:
            msg (Dict): Messsag to publish
            topic_id (str): Id of the topic

        Returns:
            str: Id of the published msg
        """
        topic_path = self.publisher.topic_path(self.project_id, topic_id)

        # Msg must be a bytestring

        msg = json.dumps(msg, cls=UUIDEncoder).encode("utf-8")

        # Wait for the returned future
        future = self.publisher.publish(topic_path, msg)
        return future.result()

    def subscribe(
        self: "GCPMessenger",
        callback: Callable,
        sub_id: str,
        timeout: Optional[float] = None,
    ) -> None:
        """Subscribe to a topic.

        Args:
            callback (Callable): Callback to invoke when a msg is received
            sub_id (str): Id of the topic.
            timeout (Union[float, None]): Time to wait for msgs. Defaults to None. # noqa: E501
        """
        subscription_path = self.subscriber.subscription_path(self.project_id, sub_id)

        streaming_pull_future = self.subscriber.subscribe(
            subscription_path,
            callback=callback,
            await_callbacks_on_shutdown=True,
        )

        with self.subscriber:
            try:
                streaming_pull_future.result(timeout=timeout)
            except TimeoutError as e:
                logger.error(
                    f"GCPMessenger timeout error while pulling messages. Error: {e}"
                )
                streaming_pull_future.cancel()

    def parse_messenger_msg(self: "GCPMessenger", messenger_msg: GCPMessage) -> Message:
        """Parse messenger msg to blackcap mesage schema.

        Args:
            messenger_msg (GCPMessage): GCPMessenger message

        Returns:
            Message: Parsed Message
        """
        return Message.parse_raw(messenger_msg.data)

    def echo_msg(self: "GCPMessenger", msg: Message) -> None:
        """Echo msgs to stdout.

        Args:
            msg (Message): Message to echo
        """
        print(msg)
        msg.ack()



"""Blackcap Schedule DELETE route schemas."""

from typing import Any, Dict, List, Union

from pydantic import BaseModel
from pydantic.types import UUID4

from blackcap.schemas.schedule import Schedule


class ScheduleDelete(BaseModel):
    """Schema to parse delete schedule requests."""

    schedule_id: UUID4


class SchedulePUTRequest(BaseModel):
    """Schedule PUT request schema."""

    schedule_list: List[ScheduleDelete]


class SchedulePUTResponse(BaseModel):
    """Schedule PUT response schema."""

    items: Dict[str, List[Union[Schedule, Any]]] = {}



"""Cluster interfaces."""


from blackcap.cluster.registry import ClusterRegistry
from blackcap.cluster.slurm_cluster import SlurmCluster


cluster_registry = ClusterRegistry()
cluster_registry.add_cluster(SlurmCluster())



"""Job BLoCs."""

from typing import List

from logzero import logger
from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper
from sqlalchemy import select

from blackcap.db import DBSession
from blackcap.models.job import JobDB
from blackcap.schemas.api.job.delete import JobDelete
from blackcap.schemas.api.job.get import JobGetQueryParams, JobQueryType
from blackcap.schemas.api.job.post import JobCreate
from blackcap.schemas.api.job.put import JobUpdate
from blackcap.schemas.job import Job
from blackcap.schemas.user import User


def create_job(job_create_list: List[JobCreate], user_creds: User) -> List[Job]:
    """Create jobs in the DB from JobCreate request.

    Args:
        job_create_list (List[JobCreate]): JobCreate request
        user_creds (User): User credentials.

    Raises:
        Exception: error

    Returns:
        List[Job]: List of created clusters
    """
    with DBSession() as session:
        try:
            job_db_create_list: List[JobDB] = [
                JobDB(
                    protagonist_id=job_create.user.user_id,
                    **job_create.job.dict(),
                )
                for job_create in job_create_list
            ]
            JobDB.bulk_create(job_db_create_list, session)
            return [Job(job_id=obj.id, **obj.to_dict()) for obj in job_db_create_list]
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to create jobs: {e}")
            raise e


def get_job(  # noqa: C901
    query_params: JobGetQueryParams, user_creds: User
) -> List[Job]:
    """Query DB for jobs.

    Args:
        query_params (JobGetQueryParams): Query params from request
        user_creds (User): User credentials.

    Raises:
        e: error
        Exception: error

    Returns:
        List[Job]: List of jobs returned from DB
    """
    job_list: List[JobDB] = []

    stmt = ""

    if query_params.query_type == JobQueryType.GET_ALL_JOBS:
        stmt = select(JobDB).where(JobDB.protagonist_id == user_creds.user_id)
    if query_params.query_type == JobQueryType.GET_JOBS_BY_ID:
        if query_params.job_id is None:
            e = ValidationError(
                errors=[
                    ErrorWrapper(ValueError("field required"), "job_id"),
                ],
                model=JobGetQueryParams,
            )
            raise e
        stmt = (
            select(JobDB)
            .where(JobDB.protagonist_id == user_creds.user_id)
            .where(JobDB.id == query_params.job_id)
        )
    if query_params.query_type == JobQueryType.GET_JOBS_BY_CLUSTER_ID:
        if query_params.cluster_id is None:
            e = ValidationError(
                errors=[
                    ErrorWrapper(ValueError("field required"), "cluster_id"),
                ],
                model=JobGetQueryParams,
            )
            raise e
        stmt = (
            select(JobDB)
            .where(JobDB.protagonist_id == user_creds.user_id)
            .where(JobDB.id == query_params.cluster_id)
        )
    if query_params.query_type == JobQueryType.GET_JOBS_BY_PROTAGONIST_ID:
        stmt = select(JobDB).where(JobDB.protagonist_id == user_creds.user_id)
    if query_params.query_type == JobQueryType.GET_JOBS_BY_STATUS:
        if query_params.job_status is None:
            e = ValidationError(
                errors=[
                    ErrorWrapper(ValueError("field required"), "job_status"),
                ],
                model=JobGetQueryParams,
            )
            raise e
        stmt = (
            select(JobDB)
            .where(JobDB.protagonist_id == user_creds.user_id)
            .where(JobDB.id == query_params.job_status)
        )

    with DBSession() as session:
        try:
            job_db_list: List[JobDB] = session.execute(stmt).scalars().all()
            job_list = [Job(job_id=obj.id, **obj.to_dict()) for obj in job_db_list]
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to fetch jobs due to {e}")
            raise e

    return job_list


def update_job(job_update_list: List[JobUpdate], user_creds: User) -> List[Job]:
    """Update job in the DB from JobUpdate request.

    Args:
        job_update_list (List[JobUpdate]): List of JobUpdate request
        user_creds (User): User credentials.

    Raises:
        Exception: error

    Returns:
        List[Job]: List of Instance of Updated Job
    """
    stmt = (
        select(JobDB)
        .where(JobDB.protagonist_id == user_creds.user_id)
        .where(JobDB.id.in_([job_update.template_id for job_update in job_update_list]))
    )
    with DBSession() as session:
        try:
            job_db_update_list: List[JobDB] = session.execute(stmt).scalars().all()
            updated_job_list = []
            for job in job_db_update_list:
                for job_update in job_update_list:
                    if job_update.job_id == job.id:
                        job_update_dict = job_update.dict(exclude_defaults=True)
                        job_update_dict.pop("job_id")
                        updated_job = job.update(session, **job_update_dict)
                        updated_job_list.append(
                            Job(
                                job_id=updated_job.id,
                                **updated_job.to_dict(),
                            )
                        )
            return updated_job_list
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to update job: {job.to_dict()} due to {e}")
            raise e


def delete_job(job_delete_list: List[JobDelete], user_creds: User) -> List[Job]:
    """Delete job in the DB from JobDelete request.

    Args:
        job_delete_list (List[JobDelete]): List of JobDelete request
        user_creds (User): User credentials.

    Raises:
        Exception: error

    Returns:
        List[Job]: List of Instance of Deleted Job
    """
    stmt = (
        select(JobDB)
        .where(JobDB.protagonist_id == user_creds.user_id)
        .where(JobDB.id.in_([job.job_id for job in job_delete_list]))
    )
    with DBSession() as session:
        try:
            job_db_delete_list: List[JobDB] = session.execute(stmt).scalars().all()
            deleted_job_list = []
            for job in job_db_delete_list:
                job.delete(session)
                deleted_job_list.append(Job(job_id=job.id, **job.to_dict()))
            return deleted_job_list
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to delete job: {job.to_dict()} due to {e}")
            raise e



"""Blackcap User GET route schemas."""

from enum import Enum, unique
from typing import List, Optional

from pydantic import BaseModel

from blackcap.schemas.api.common import ResponseSchema
from blackcap.schemas.user import User
from pydantic.types import UUID4


@unique
class UserQueryType(Enum):
    """User Query type enum."""

    GET_ALL_USERS = "get_all_users"
    GET_USER_BY_ID = "get_user_by_id"
    GET_USERS_BY_EMAIL = "get_users_by_email"
    GET_USERS_BY_ORGANISATION = "get_users_by_organisation"


class UserGetQueryParams(BaseModel):
    """User GET request query params schema."""

    query_type: UserQueryType
    user_id: Optional[UUID4]
    email: Optional[str]
    organisation: Optional[str]


class UserGetResponse(ResponseSchema):
    """User GET response schema."""

    items: List[User] = []



"""Flow steps."""

from dataclasses import dataclass
from typing import Callable, Dict, List


@dataclass
class Prop:
    """Parameter for a step."""

    data: Dict
    description: str


@dataclass
class FuncProp:
    """Function parameter for a step."""

    func: Callable
    params: Dict
    description: str


@dataclass
class Step:
    """A Step of the flow."""

    forward_call: Callable[[List[Prop]], List[Prop]]
    backward_call: Callable[[List[Prop]], List[Prop]]


def dummy_backward(prop: List[Prop]) -> Prop:
    """Dummy backward step function."""
    return prop



"""Blackcap Cluster DELETE route schemas."""

from typing import Any, Dict, List, Union

from pydantic import BaseModel
from pydantic.types import UUID4


from blackcap.schemas.api.common import ResponseSchema
from blackcap.schemas.cluster import Cluster


class ClusterDelete(BaseModel):
    """Schema to parse delete cluster requests."""

    cluster_id: UUID4


class ClusterPUTRequest(BaseModel):
    """Cluster PUT request schema."""

    cluster_list: List[ClusterDelete]


class ClusterPUTResponse(ResponseSchema):
    """Cluster PUT response schema."""

    items: Dict[str, List[Union[Cluster, Any]]] = {}



"""Blackcap Job POST route schemas."""

from typing import Any, Dict, List, Optional, Union


from pydantic import BaseModel

from blackcap.schemas.job import Job


class JobCreate(BaseModel):
    """Schema to parse create job requests."""

    name: str
    description: Optional[str]
    job_type: str = ""
    specification: Dict = {}
    job_metadata: Dict = {}
    script: Optional[str]


class JobPOSTRequest(BaseModel):
    """Job POST request schema."""

    job_list: List[JobCreate]


class JobPOSTResponse(BaseModel):
    """Job POST response schema."""

    items: Dict[str, List[Union[Job, Any]]] = {}



"""Base cluster inteface."""

from abc import ABC, abstractmethod
from typing import Any, List

from blackcap.configs import config_registry
from blackcap.messenger import messenger_registry
from blackcap.schemas.schedule import Schedule

config = config_registry.get_config()
messenger = messenger_registry.get_messenger(config.MESSENGER)


class BaseCluster(ABC):
    """Base cluster interface."""

    CONFIG_KEY = "CLUSTER"
    CONFIG_KEY_DEF_VAL = "ARGO"

    # Change this value in custom auther implementations.
    CONFIG_KEY_VAL = "ARGO"

    @abstractmethod
    def prepare_job(self: "BaseCluster", schedule: Schedule) -> None:
        """Prepare job for submission.

        Args:
            schedule (Schedule): Schedule Object
        """

    @abstractmethod
    def submit_job(self: "BaseCluster", schedule: Schedule) -> str:
        """Submit job to the cluster.

        Args:
            schedule (Schedule): Schedule Object

        Returns:
            str: Job ID
        """
        pass

    @abstractmethod
    def get_job_status(self: "BaseCluster", job_id: str) -> List[str]:
        """Get status of a job.

        Args:
            job_id (str): ID of the job

        Returns:
            List[str]: List of  status of the jobs
        """
        pass

    def process_schedule_msg(self: "BaseCluster", messenger_msg: Any) -> None:
        """Submit job to the cluster.

        Args:
            messenger_msg (Any): message in Messenger specific format
        """
        schedule = messenger.parse_messenger_msg(messenger_msg)
        self.prepare_job(schedule)
        self.submit_job(schedule)



"""Blackcap Cluster POST route schemas."""

from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel

from blackcap.schemas.api.common import ResponseSchema
from blackcap.schemas.cluster import Cluster


class ClusterCreate(BaseModel):
    """Schema to parse create Cluster requests."""

    name: str
    cluster_type: str
    status: str = "ONLINE"
    cluster_caps: Optional[str]
    messenger: str
    messenger_queue: str


class ClusterPOSTRequest(BaseModel):
    """Cluster POST request schema."""

    cluster_list: List[ClusterCreate]


class ClusterPOSTResponse(ResponseSchema):
    """Cluster POST response schema."""

    items: Dict[str, List[Union[Cluster, Any]]] = {}



"""Schedule DBModel."""

from sqlalchemy import Column, DateTime, String

from blackcap.models.meta.mixins import (
    DBModel,
    SurrogatePKUUID,
    TimestampMixin,
)
from blackcap.models.meta.orm import reference_col


class ScheduleDB(DBModel, TimestampMixin, SurrogatePKUUID):
    """Schedule table."""

    __tablename__ = "schedule"
    job_id = reference_col("job")
    assigned_cluster_id = reference_col("cluster")
    # TODO: this should be an Enum with the possible status values
    status = Column(String, nullable=False, default="PENDING")
    started_at = Column(DateTime(timezone=True), nullable=True)
    finished_at = Column(DateTime(timezone=True), nullable=True)
    protagonist_id = reference_col("protagonist")



"""Auth utils."""

from functools import wraps
from http import HTTPStatus
import sys
from typing import Callable

import click
from flask import make_response, request

from blackcap.auther import auther_registry
from blackcap.configs import ComposeDefaultConfig, config_registry
from blackcap.schemas.api.common import ResponseSchema
from blackcap.schemas.user import User

config = config_registry.get_config()


def check_auth(config: ComposeDefaultConfig) -> User:
    """Check authorization in CLI.

    Args:
        config (ComposeDefaultConfig): config object

    Returns:
        User: User object
    """
    user_access_token = config.USER_ACCESS_TOKEN
    auther = auther_registry.get_auther(config.AUTHER)
    try:
        user = auther.extract_user_from_token(user_access_token)
    except Exception:
        click.secho("Authorization failed\n\n Exiting....")
        sys.exit(1)
    if user is None:
        click.secho("Authorization failed\n\n Exiting....")
        sys.exit(1)
    return user


def check_authentication(flask_route: Callable) -> Callable:
    """Decorator to check authentication in flask routes.

    Args:
        flask_route (Callable): Flask route

    Returns:
        Callable: Flask route
    """

    @wraps(flask_route)
    def wrapper() -> Callable:
        """Flask route wrapper.

        Returns:
            Callable: Flask route or Authentication error
        """
        # Extract token from cookie
        # Cookie example:
        # {"imcloud": "Bearer ey0038295.8kjfdfkd.kjfslusnv"}
        try:
            token_from_cookie = request.cookies.get("imcloud").split(" ")[1]
            # Try to extract user from token
            user = auther_registry.get_auther(config.AUTHER).extract_user_from_token(
                token_from_cookie
            )
            # Return Authentication error if user is none
            if user is None:
                raise Exception("User not found")
        except Exception:
            response_body = ResponseSchema(
                msg="Authentication failed. Please login again.",
                errors={
                    "main": [
                        {
                            "loc": "creds",
                            "msg": "unauthorized",
                            "type": "authentication",
                        }
                    ]
                },
            )
            return make_response(response_body.json(), HTTPStatus.UNAUTHORIZED)

        # Return flask route if auth is successful
        return flask_route(user)

    return wrapper



"""Flow executor."""

from dataclasses import dataclass
from typing import Dict

from blackcap.flow.flow import Flow, FlowExecError, FlowStatus, get_outer_function
from blackcap.flow.step import FuncProp, Prop


@dataclass
class Executor:
    """Flow executor."""

    flow: Flow
    # TODO: Strongly type it in the future
    config: Dict

    def run(self: "Executor") -> Flow:  # noqa: C901
        """Execute flow."""
        # Set Flow status to executing
        self.flow.status = FlowStatus.EXECUTING
        for index, step in enumerate(self.flow.steps):
            try:
                # Prepare function inputs
                functions_inputs = []
                # Check the prop type
                for item in self.flow.inputs[index]:
                    if type(item) == FuncProp:
                        try:
                            input_list_from_func = item.func(**item.params)
                        except Exception as e:
                            raise FlowExecError(
                                human_description="Invoking FuncProp failed",
                                error=e,
                                error_type=type(e),
                                is_user_facing=False,
                                error_in_function=item.func,
                                step_index=index,
                            ) from e
                        functions_inputs += input_list_from_func
                    elif type(item) == Prop:
                        functions_inputs.append(item)
                    else:
                        raise FlowExecError(
                            human_description="Unkown type of input is used to invoke a step function",
                            error=f"Input type: {type(self.flow.inputs[index])} is not recognised",
                            error_type="Unkown input error",
                            is_user_facing=False,
                            error_in_function=get_outer_function(),
                        )
                # Replace current function input with the prepared inputs
                self.flow.inputs[index] = functions_inputs

                # Invoke function with prepared inputs
                forward_out = step.forward_call(self.flow.inputs[index])
                self.flow.forward_outputs.append(forward_out)
            except Exception as e:
                # TODO: Add logging for failed forward calls
                # Set flow status to failed and append error
                self.flow.status = FlowStatus.FAILED
                flow_error = FlowExecError(
                    human_description="A forward step function failed",
                    error=e,
                    error_type=type(e),
                    is_user_facing=False,
                    error_in_function=get_outer_function(),
                )
                flow_error.step_index = index
                self.flow.errors.append(flow_error)
                for back_index in reversed(range(0, index)):
                    try:
                        backward_out = self.flow.steps[back_index].backward_call(
                            self.flow.inputs[back_index]
                            + self.flow.forward_outputs[back_index],
                        )
                        self.flow.backward_outputs.append(backward_out)
                    except Exception as e:
                        # TODO: Add central logging here
                        flow_error = FlowExecError(
                            human_description="A backward step function failed",
                            error=e,
                            error_type=type(e),
                            is_user_facing=False,
                            error_in_function=get_outer_function(),
                        )
                        flow_error.step_index = index
                        self.flow.errors.append(flow_error)
                return self.flow
        self.flow.status = FlowStatus.PASSED
        return self.flow



"""Blackcap Cluster PUT route schemas."""

from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel
from pydantic.types import UUID4

from blackcap.schemas.api.common import ResponseSchema
from blackcap.schemas.cluster import Cluster


class ClusterUpdate(BaseModel):
    """Schema to parse update cluster requests."""

    cluster_id: UUID4
    cluster_type: Optional[str]
    cluster_caps: Optional[str]
    messenger: Optional[str]
    messenger_queue: Optional[str]


class ClusterPUTRequest(BaseModel):
    """Cluster PUT request schema."""

    cluster_list: List[ClusterUpdate]


class ClusterPUTResponse(ResponseSchema):
    """Cluster PUT response schema."""

    items: Dict[str, List[Union[Cluster, Any]]] = {}



"""Cluster registry."""

from typing import Optional

from blackcap.cluster.base import BaseCluster


class ClusterRegistry:
    """Cluster registry."""

    clusters = {}

    def add_cluster(self: "ClusterRegistry", cluster: BaseCluster) -> None:
        """Add custom cluster to registry.

        Args:
            cluster (BaseCluster): Custom cluster implementation
        """
        self.clusters[cluster.CONFIG_KEY_VAL] = cluster

    def get_cluster(self: "ClusterRegistry", cluster: str) -> Optional[BaseCluster]:
        """Get cluster.

        Args:
            cluster (str): Cluster name

        Returns:
            Optional[BaseCluster]: Returns the cluster if found else None
        """
        return self.clusters.get(cluster)



"""Blackcap Schedule POST route schemas."""

from typing import Any, Dict, List, Optional, Union

from blackcap.schemas.schedule import Schedule

from pydantic import BaseModel, UUID4


class ScheduleCreate(BaseModel):
    """Schema to parse create schedule requests."""

    job_id: UUID4
    assigned_cluster_id: Optional[UUID4]
    messenger: Optional[str]
    messenger_queue: Optional[str]
    status: str = "PENDING"


class SchedulePOSTRequest(BaseModel):
    """Schedule POST request schema."""

    schedule_list: List[ScheduleCreate]


class SchedulePOSTResponse(BaseModel):
    """Schedule POST response schema."""

    items: Dict[str, List[Union[Schedule, Any]]] = {}



"""Blackcap User DELETE route schemas."""

from typing import Any, Dict, List, Union

from pydantic import BaseModel
from pydantic.types import UUID4

from blackcap.schemas.api.common import ResponseSchema
from blackcap.schemas.user import User


class UserDelete(BaseModel):
    """Schema to parse delete user requests."""

    user_id: UUID4


class UserDELETEResponse(ResponseSchema):
    """User DELETE response schema."""

    items: Dict[str, List[Union[User, Any]]] = {}



"""Blackcap cluster routes."""

from flask import Blueprint


cluster_bp = Blueprint("cluster", __name__, url_prefix="/v1/cluster")


from blackcap.routes.cluster.get import get_cluster  # noqa: F401, E402, I100
from blackcap.routes.cluster.post import post_cluster  # noqa: F401, E402, I100



"""Schedule schema."""

from pydantic import BaseModel
from pydantic.types import UUID4

from blackcap.schemas.job import Job


class Schedule(BaseModel):
    """Schedule schema."""

    schedule_id: UUID4
    job_id: UUID4
    job: Job
    assigned_cluster_id: UUID4
    messenger: str
    messenger_queue: str
    status: str = "PENDING"



"""Blackcap job POST route."""

from http import HTTPStatus
import json

from flask import make_response, request, Response
from pydantic import parse_obj_as, ValidationError
from sqlalchemy.exc import SQLAlchemyError

from blackcap.blocs.cluster import create_cluster
from blackcap.routes.cluster import cluster_bp
from blackcap.schemas.api.cluster.post import ClusterPOSTRequest, ClusterPOSTResponse
from blackcap.schemas.user import User
from blackcap.utils.auth import check_authentication


@cluster_bp.post("/")
@check_authentication
def post(user: User) -> Response:
    """Post cluster.

    Args:
        user (User): Extracted user from request

    Returns:
        Response: Flask response
    """
    # Parse json from request
    try:
        cluster_create_request_list = parse_obj_as(
            ClusterPOSTRequest, json.loads(request.data)
        ).cluster_list
    except ValidationError as e:
        response_body = ClusterPOSTResponse(
            msg="json validation failed", errors={"main": e.errors()}
        )
        return make_response(response_body.json(), HTTPStatus.BAD_REQUEST)
    except Exception:
        response_body = ClusterPOSTResponse(
            msg="unknown error", errors={"main": ["unknown internal error"]}
        )
        return make_response(response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR)

    # ceate cluster in DB and publish msg
    try:
        cluster_list = create_cluster(cluster_create_request_list, user)
    except SQLAlchemyError:
        response_body = ClusterPOSTResponse(
            msg="internal databse error", errors={"main": ["unknown internal error"]}
        )
        return make_response(response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR)
    except Exception:
        response_body = ClusterPOSTResponse(
            msg="unknown error", errors={"main": ["unknown internal error"]}
        )
        return make_response(response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR)

    # return created cluster in response
    response_body = ClusterPOSTResponse(
        msg="cluster successfully created", items={"cluster_list": cluster_list}
    )
    return make_response(response_body.json(), HTTPStatus.OK)



"""Job schema."""

from typing import Dict, Optional

from pydantic import BaseModel
from pydantic.types import UUID4


class Job(BaseModel):
    """Job schema."""

    job_id: UUID4
    name: str
    description: str = ""
    job_type: str = ""
    specification: Dict = {}
    job_metadata: Dict = {}
    script: Optional[str]



"""Blackcap schedule POST route."""

from http import HTTPStatus


from flask import make_response, request, Response
from pydantic import ValidationError
from sqlalchemy.exc import SQLAlchemyError

from blackcap.blocs.schedule import create_schedule
from blackcap.routes.schedule import schedule_bp
from blackcap.schemas.api.schedule.post import ScheduleCreate, SchedulePOSTResponse
from blackcap.schemas.user import User
from blackcap.utils.auth import check_authentication


@schedule_bp.post("/")
@check_authentication
def post_schedule(user: User) -> Response:
    """Post schedule.

    Args:
        user (User): Extracted user from request

    Returns:
        Response: Flask response
    """
    # Parse json from request
    try:
        schedule_create_request_list = ScheduleCreate.parse_obj(request.json)
    except ValidationError as e:
        response_body = SchedulePOSTResponse(
            msg="json validation failed", errors={"main": e.errors()}
        )
        return make_response(response_body.json(), HTTPStatus.BAD_REQUEST)
    except Exception:
        response_body = SchedulePOSTResponse(
            msg="unknown error", errors={"main": ["unknown internal error"]}
        )
        return make_response(response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR)

    # ceate schedule in DB and publish msg
    try:
        schedule_list = create_schedule(schedule_create_request_list, user)
    except SQLAlchemyError:
        response_body = SchedulePOSTResponse(
            msg="internal databse error", errors={"main": ["unknown internal error"]}
        )
        return make_response(response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR)
    except Exception:
        response_body = SchedulePOSTResponse(
            msg="unknown error", errors={"main": ["unknown internal error"]}
        )
        return make_response(response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR)

    # return created schedule in response
    response_body = SchedulePOSTResponse(
        msg="schedule successfully created", items={"schedule_list": schedule_list}
    )
    return make_response(response_body.json(), HTTPStatus.OK)



"""Blackcap Schedule PUT route schemas."""

from typing import Any, Dict, List, Union

from pydantic import BaseModel
from pydantic.types import UUID4

from blackcap.schemas.schedule import Schedule


class ScheduleUpdate(BaseModel):
    """Schema to parse update schedule requests."""

    schedule_id: UUID4


class SchedulePUTRequest(BaseModel):
    """Schedule PUT request schema."""

    schedule_list: List[ScheduleUpdate]


class SchedulePUTResponse(BaseModel):
    """Schedule PUT response schema."""

    items: Dict[str, List[Union[Schedule, Any]]] = {}



"""Mixins for database models."""

from datetime import datetime
from typing import Any, List, Optional, Union
import uuid

from sqlalchemy import Column, DateTime, delete, Integer
from sqlalchemy.orm import declarative_base
from sqlalchemy.orm.session import Session
from sqlalchemy_serializer import SerializerMixin

from blackcap.models.meta.helpers import GUID

# declarative base class
Base = declarative_base()


class TimestampMixin(object):
    """Mixin that adds timestamp."""

    created_at = Column(
        DateTime(timezone=True), nullable=False, default=datetime.utcnow
    )
    updated_at = Column(
        DateTime(timezone=True), default=datetime.utcnow, onupdate=datetime.utcnow
    )


class CRUDMixin(object):
    """Mixin that adds convenience methods for CRUD operations."""

    @classmethod
    def create(cls: Any, session: Session, **kwargs: Any) -> Any:
        """Create a new record and save it to the database.

        Args:
            session: (Session): database session to use
            **kwargs (Any): kwargs to pass to cls for init

        Returns:
            Any: Instance of self
        """
        instance = cls(**kwargs)
        return instance.save(session)

    def save(self: "CRUDMixin", session: Session, commit: bool = True) -> Any:
        """Save the record.

        Args:
            session: (Session): database session to use
            commit (bool): Flag to control commit behaviour. Defaults to True. # noqa: E501

        Returns:
            Any: Instance of self
        """
        session.add(self)
        if commit:
            session.commit()
        return self

    @classmethod
    def bulk_create(cls: Any, obj_list: List[Any], session: Session) -> Any:
        """Bulk create new records and save them to the database.

        Args:
            obj_list (List[Any]): List of self
            session: (Session): database session to use

        Returns:
            Any: True if scuccess
        """
        session.add_all(obj_list)
        session.commit()
        return True

    def update(
        self: "CRUDMixin", session: Session, commit: bool = True, **kwargs: Any
    ) -> Any:
        """Update specific fields of a record.

        Args:
            session: (Session): database session to use
            commit (bool): Flag to control commit behaviour. Defaults to True.
            **kwargs (Any): kwargs to update cls params

        Returns:
            Any: Instance of self
        """
        for attr, value in kwargs.items():
            setattr(self, attr, value)
        return commit and self.save(session) or self

    @classmethod
    def bulk_update(cls: Any, obj_list: List[Any], session: Session) -> Any:
        """Bulk update records and save them to the database.

        Args:
            obj_list (List[Any]): List of instances of self
            session: (Session): database session to use

        Returns:
            Any: List of updated instances self
        """
        instance_list: List[Any] = []
        for item in obj_list:
            instance_list.append(cls(**item))

        session.add_all(instance_list)
        session.commit()
        return instance_list

    def delete(
        self: "CRUDMixin", session: Session, commit: bool = True
    ) -> Union[bool, Any]:
        """Remove the record from the database.

        Args:
            session: (Session): database session to use
            commit (bool): Flag to control commit behaviour. Defaults to True. # noqa: E501

        Returns:
            Union[bool, Any]: True if commit is successful
        """
        session.delete(self)
        return commit and session.commit()

    def bulk_delete(cls: Any, obj_list: List[Any], session: Session) -> Any:
        """Bulk delete records from the database.

        Args:
            obj_list (List[Any]): List of self
            session: (Session): database session to use

        Returns:
            Any: True if scuccess
        """
        session.execute(delete(cls).where(cls.id._in([obj.id for obj in obj_list])))
        session.commit()
        return True


class DBModel(CRUDMixin, SerializerMixin, Base):
    """Base model class that includes CRUD convenience methods."""

    __abstract__ = True


# From Mike Bayer's "Building the app" talk
# https://speakerdeck.com/zzzeek/building-the-app
class SurrogatePK(object):
    """Adds a surrogate int 'primary key' column named ``id`` to any orm class."""  # noqa: E501

    __table_args__ = {"extend_existing": True}

    id = Column(Integer, primary_key=True)

    @classmethod
    def get_by_id(cls: Any, record_id: Union[str, int]) -> Union[None, Any]:
        """Get record by ID.

        Args:
            record_id (Union[str, int]): Id of the record to fetch

        Returns:
            Union[None, Any]: Return the instance of self if found or else None
        """
        if any(
            (
                isinstance(record_id, str) and record_id.isdigit(),
                isinstance(record_id, (int, float)),
            )
        ):
            return cls.query.get(int(record_id))
        return None


class SurrogatePKUUID(object):
    """Adds a surrogate int 'primary key' column named ``id`` to any orm class."""  # noqa: B950

    __table_args__ = {"extend_existing": True}

    id = Column(GUID, primary_key=True, default=lambda: str(uuid.uuid4()))

    @classmethod
    def get_by_uuid(cls: Any, record_id: Union[str, uuid.UUID]) -> Optional[Any]:
        """Get record by ID.

        Args:
            record_id (Union[str, uuid.UUID]): Id of the record to fetch

        Returns:
            Union[None, Any]: Return the instance of self if found or else None
        """
        if any((isinstance(record_id, str), isinstance(record_id, uuid.UUID))):
            return cls.query.get(uuid.UUID(str(record_id)))
        return None



"""Blackcap Job DELETE route schemas."""

from typing import Any, Dict, List, Union

from pydantic import BaseModel, UUID4

from blackcap.schemas.job import Job


class JobDelete(BaseModel):
    """Schema to parse delete job requests."""

    job_id: UUID4


class JobDELETERequest(BaseModel):
    """Job DELETE request schema."""

    job_list: List[JobDelete]


class JobDELETEResponse(BaseModel):
    """Job DELETE response schema."""

    items: Dict[str, List[Union[Job, Any]]] = {}



"""Blackcap job POST route."""

from http import HTTPStatus
import json

from flask import make_response, request, Response
from pydantic import parse_obj_as, ValidationError
from sqlalchemy.exc import SQLAlchemyError

from blackcap.blocs.job import create_job
from blackcap.routes.job import job_bp
from blackcap.schemas.api.job.post import JobPOSTRequest, JobPOSTResponse
from blackcap.schemas.user import User
from blackcap.utils.auth import check_authentication


@job_bp.post("/")
@check_authentication
def post_job(user: User) -> Response:
    """Post job.

    Args:
        user (User): Extracted user from request

    Returns:
        Response: Flask response
    """
    # Parse json from request
    try:
        job_create_request_list = parse_obj_as(
            JobPOSTRequest, json.loads(request.data)
        ).job_list
    except ValidationError as e:
        response_body = JobPOSTResponse(
            msg="json validation failed", errors={"main": e.errors()}
        )
        return make_response(response_body.json(), HTTPStatus.BAD_REQUEST)
    except Exception:
        response_body = JobPOSTResponse(
            msg="unknown error", errors={"main": ["unknown internal error"]}
        )
        return make_response(response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR)

    # ceate job in DB and publish msg
    try:
        job_list = create_job(job_create_request_list, user)
    except SQLAlchemyError:
        response_body = JobPOSTResponse(
            msg="internal databse error", errors={"main": ["unknown internal error"]}
        )
        return make_response(response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR)
    except Exception:
        response_body = JobPOSTResponse(
            msg="unknown error", errors={"main": ["unknown internal error"]}
        )
        return make_response(response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR)

    # return created job in response
    response_body = JobPOSTResponse(
        msg="job successfully created", items={"job_list": job_list}
    )
    return make_response(response_body.json(), HTTPStatus.OK)



"""Cluster BLoCs."""

from typing import List

from logzero import logger
from sqlalchemy import select
from sqlalchemy.exc import SQLAlchemyError

from blackcap.db import DBSession
from blackcap.flow import Flow, FlowExecError, get_outer_function, Prop, Step
from blackcap.models.cluster import ClusterDB
from blackcap.schemas.api.cluster.delete import ClusterDelete
from blackcap.schemas.api.cluster.get import (
    ClusterGetQueryParams,
    ClusterQueryType,
)
from blackcap.schemas.api.cluster.post import ClusterCreate
from blackcap.schemas.api.cluster.put import ClusterUpdate
from blackcap.schemas.cluster import Cluster
from blackcap.schemas.user import User


def create_cluster(
    cluster_create_list: List[ClusterCreate], user_creds: User
) -> List[Cluster]:
    """Create clusters from create cluster request.

    Args:
        cluster_create_list (List[ClusterCreate]): ClusterCreate request list
        user_creds (User): User credentials.

    Raises:
        Exception: error

    Returns:
        List[Cluster]: List of created clusters
    """
    with DBSession() as session:
        try:
            cluster_db_create_list: List[ClusterDB] = [
                ClusterDB(**cluster_create.cluster.dict())
                for cluster_create in cluster_create_list
            ]
            ClusterDB.bulk_create(cluster_db_create_list, session)
            return [
                Cluster(cluster_id=obj.id, **obj.to_dict())
                for obj in cluster_db_create_list
            ]
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to create clusters: {e}")
            raise e


def get_clusters(
    query_params: ClusterGetQueryParams, user_creds: User
) -> List[Cluster]:
    """Query DB for clusters.

    Args:
        query_params (ClusterGetQueryParams): Query params from request
        user_creds (User): User credentials.

    Raises:
        Exception: error

    Returns:
        List[Cluster]: List of clusters returned from DB
    """
    cluster_list: List[ClusterDB] = []

    stmt = ""

    if query_params.query_type == ClusterQueryType.GET_ALL_CLUSTER:
        stmt = select(ClusterDB)

    with DBSession() as session:
        try:
            cluster_list: List[ClusterDB] = session.execute(stmt).scalars().all()
            return [Cluster(cluster_id=obj.id, **obj.to_dict()) for obj in cluster_list]
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to fetch clusters due to {e}")
            raise e


def update_cluster(
    cluster_update_list: List[ClusterUpdate], user_creds: User
) -> List[Cluster]:
    """Update cluster in the DB from ClusterUpdate request.

    Args:
        cluster_update_list (List[ClusterUpdate]): List of ClusterUpdate request
        user_creds (User): User credentials.

    Raises:
        Exception: error

    Returns:
        List[Cluster]: List of Instance of Updated Cluster
    """
    stmt = (
        select(ClusterDB)
        .where(ClusterDB.protagonist_id == user_creds.user_id)
        .where(
            ClusterDB.id.in_(
                [cluster_update.template_id for cluster_update in cluster_update_list]
            )
        )
    )
    with DBSession() as session:
        try:
            cluster_db_update_list: List[ClusterDB] = (
                session.execute(stmt).scalars().all()
            )
            updated_cluster_list = []
            for cluster in cluster_db_update_list:
                for cluster_update in cluster_update_list:
                    if cluster_update.cluster_id == cluster.id:
                        cluster_update_dict = cluster_update.dict(exclude_defaults=True)
                        cluster_update_dict.pop("cluster_id")
                        updated_cluster = cluster.update(session, **cluster_update_dict)
                        updated_cluster_list.append(
                            Cluster(
                                cluster_id=updated_cluster.id,
                                **updated_cluster.to_dict(),
                            )
                        )
            return updated_cluster_list
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to update cluster: {cluster.to_dict()} due to {e}")
            raise e


def delete_cluster(
    cluster_delete_list: List[ClusterDelete], user_creds: User
) -> List[Cluster]:
    """Delete cluster in the DB from ClusterDelete request.

    Args:
        cluster_delete_list (List[ClusterDelete]): List of ClusterDelete request
        user_creds (User): User credentials.

    Raises:
        Exception: error

    Returns:
        List[Cluster]: List of Instance of Deleted Cluster
    """
    stmt = (
        select(ClusterDB)
        .where(ClusterDB.protagonist_id == user_creds.user_id)
        .where(
            ClusterDB.id.in_([cluster.cluster_id for cluster in cluster_delete_list])
        )
    )
    with DBSession() as session:
        try:
            cluster_db_delete_list: List[ClusterDB] = (
                session.execute(stmt).scalars().all()
            )
            deleted_cluster_list = []
            for cluster in cluster_db_delete_list:
                cluster.delete(session)
                deleted_cluster_list.append(
                    Cluster(cluster_id=cluster.id, **cluster.to_dict())
                )
            return deleted_cluster_list
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to delete cluster: {cluster.to_dict()} due to {e}")
            raise e


###
# Flow BLoCs
###


def create_cluster_db_entry(inputs: List[Prop]) -> List[Prop]:
    """Create cluster db entry step function.

    Args:
        inputs (List[Prop]):
            Expects
                0: cluster_create_request_list
                    Prop(data=cluster_create_request_list, description="List of create cluster objects")
                1: user
                    Prop(data=user, description="User")

    Raises:
        FlowExecError: Flow execution failed

    Returns:
        List[Prop]:
            Created cluster objects

            Prop(data=created_cluster_list, description="List of created cluster Objects")
    """
    try:
        cluster_create_request_list: List[ClusterCreate] = inputs[0].data
        user: User = inputs[1].data
    except Exception as e:
        raise FlowExecError(
            human_description="Parsing inputs failed",
            error=e,
            error_type=type(e),
            is_user_facing=True,
            error_in_function=get_outer_function(),
        ) from e

    try:
        created_cluster_list = create_cluster(cluster_create_request_list, user)
    except SQLAlchemyError as e:
        raise FlowExecError(
            human_description="Creating DB object failed",
            error=e,
            error_type=type(e),
            is_user_facing=False,
            error_in_function=get_outer_function(),
        ) from e
    except Exception as e:
        raise FlowExecError(
            human_description="Something bad happened",
            error=e,
            error_type=type(e),
            is_user_facing=False,
            error_in_function=get_outer_function(),
        ) from e

    return [
        Prop(data=created_cluster_list, description="List of created cluster Objects")
    ]


def revert_cluster_db_entry(inputs: List[Prop]) -> List[Prop]:
    """Delete db entry step function.

    Args:
        inputs (List[Prop]):
            Expects
                0: cluster_create_request_list
                    Prop(data=cluster_create_request_list, description="List of create cluster objects")
                1: user
                    Prop(data=user, description="User")
                2: created_cluster_list
                    Prop(data=created_data_list, description="List of created cluster objects")

    Raises:
        FlowExecError: Flow execution failed

    Returns:
        List[Prop]:
            Deleted data objects

            Prop(data=deleted_cluster_list, description="List of deleted cluster Objects")
    """
    try:
        created_cluster_list: List[Cluster] = inputs[2].data
        user: User = inputs[1].data
    except Exception as e:
        raise FlowExecError(
            human_description="Parsing inputs failed",
            error=e,
            error_type=type(e),
            is_user_facing=True,
            error_in_function=get_outer_function(),
        ) from e

    try:
        deleted_cluster_list = delete_cluster(created_cluster_list, user)
    except SQLAlchemyError as e:
        raise FlowExecError(
            human_description="Deleting DB object failed",
            error=e,
            error_type=type(e),
            is_user_facing=False,
            error_in_function=get_outer_function(),
        ) from e
    except Exception as e:
        raise FlowExecError(
            human_description="Something bad happened",
            error=e,
            error_type=type(e),
            is_user_facing=False,
            error_in_function=get_outer_function(),
        ) from e

    return [
        Prop(data=deleted_cluster_list, description="List of deleted cluster Objects")
    ]


def generate_create_cluster_flow(
    cluster_create_request_list: List[ClusterCreate], user: User
) -> Flow:
    """Generate flow for creating the cluster resource.

    Args:
        cluster_create_request_list (List[ClusterCreate]): List of cluster objects to create.
        user (User): User credentials.

    Returns:
        Flow: Create cluster flow
    """
    create_db_entry_step = Step(create_cluster_db_entry, revert_cluster_db_entry)
    # create_messenger_topic_step = Step(create_messenger_topic, delete_meesenger_topic)

    flow = Flow()

    flow.add_step(
        create_db_entry_step,
        [
            Prop(
                data=cluster_create_request_list,
                description="List of ClusterCreate Objects",
            ),
            Prop(data=user, description="User"),
        ],
    )

    return flow



"""Random Scheduler implementation of Scheduler."""

import random
from typing import List

from logzero import logger
from sqlalchemy.sql.expression import select

from blackcap.db import DBSession
from blackcap.models.cluster import ClusterDB
from blackcap.scheduler.base import BaseScheduler
from blackcap.schemas.api.schedule.post import ScheduleCreate


class RandomScheduler(BaseScheduler):
    """Random scheduler schedule jobs randomly."""

    CONFIG_KEY_VAL = "RANDOM"

    def schedule(
        self: "BaseScheduler", schedule_create: ScheduleCreate
    ) -> ScheduleCreate:
        """Create schedule from schedule request.

        Args:
            schedule_create (ScheduleCreate): Schedule create request

        Raises:
            Exception: No cluster available

        Returns:
            ScheduleCreate: Instance of Schedule Create
        """
        # fetch available clusters
        cluster_list: List[ClusterDB] = []
        with DBSession() as session:
            try:
                stmt = select(ClusterDB)
                cluster_list = session.execute(stmt).scalars().all()
            except Exception as e:
                logger.error(f"Unable to fetch clusters: {e}")
                raise e

        # Raise error if no cluster available
        if len(cluster_list) == 0:
            logger.error("No cluster available")
            raise Exception("cluster_list: List[ClusterDB]")

        # Randomly selects a cluster
        selected_cluster = random.choice(cluster_list)  # noqa: S311
        schedule_create.assigned_cluster_id = selected_cluster.id

        return schedule_create



"""Schedule BLoCs."""

from typing import List


from logzero import logger
from pydantic import ValidationError
from pydantic.error_wrappers import ErrorWrapper
from sqlalchemy import select
from sqlalchemy.exc import SQLAlchemyError


from blackcap.configs import config_registry
from blackcap.blocs.job import get_job
from blackcap.db import DBSession
from blackcap.flow import Flow, FlowExecError, FuncProp, get_outer_function, Prop, Step
from blackcap.flow.step import dummy_backward
from blackcap.messenger import messenger_registry
from blackcap.models.schedule import ScheduleDB
from blackcap.scheduler import scheduler_registry
from blackcap.schemas.api.job.get import JobGetQueryParams, JobQueryType
from blackcap.schemas.api.schedule.delete import ScheduleDelete
from blackcap.schemas.api.schedule.get import (
    ScheduleGetQueryParams,
    ScheduleQueryType,
)
from blackcap.schemas.api.schedule.post import ScheduleCreate
from blackcap.schemas.api.schedule.put import ScheduleUpdate
from blackcap.schemas.job import Job
from blackcap.schemas.message import Message, MessageType
from blackcap.schemas.schedule import Schedule
from blackcap.schemas.user import User

config = config_registry.get_config()
scheduler = scheduler_registry.get_scheduler(config.SCHEDULER)


##################
# CRUD
##################


def create_schedule(
    schedule_create_list: List[ScheduleCreate], user_creds: User
) -> List[Schedule]:
    """Create schedules and save to DB from ScheduleCreate requests.

    Args:
        schedule_create_list (List[ScheduleCreate]): List of ScheduleCreate request
        user_creds (User): User credentials.

    Raises:
        Exception: database error

    Returns:
        List[Schedule]: Instance of Schedule from DB
    """
    with DBSession() as session:
        try:
            schedule_db_create_list: List[ScheduleDB] = [
                ScheduleDB(
                    protagonist_id=user_creds.user_id,
                    **schedule.dict(),
                )
                for schedule in schedule_create_list
            ]
            ScheduleDB.bulk_create(schedule_db_create_list, session)
            return [
                Schedule(schedule_id=obj.id, **obj.to_dict())
                for obj in schedule_db_create_list
            ]
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to create schedules: {e}")
            raise e


def get_schedule(
    query_params: ScheduleGetQueryParams, user_creds: User
) -> List[Schedule]:
    """Query DB for schedules.

    Args:
        query_params (ScheduleGetQueryParams): Query params from request
        user_creds (User): User credentials.

    Raises:
        e: error
        Exception: error

    Returns:
        List[Schedule]: List of schedules returned from DB
    """
    schedule_list: List[ScheduleDB]
    if query_params.query_type == ScheduleQueryType.GET_ALL_SCHEDULES:
        stmt = select(ScheduleDB).where(ScheduleDB.protagonist_id == user_creds.user_id)
    if query_params.query_type == ScheduleQueryType.GET_SCHEDULE_BY_ID:
        if query_params.schedule_id is None:
            e = ValidationError(
                errors=[
                    ErrorWrapper(ValueError("field required"), "schedule_id"),
                ],
                model=ScheduleGetQueryParams,
            )
            raise e
        stmt = select(ScheduleDB).where(ScheduleDB.id == query_params.schedule_id)
    if query_params.query_type == ScheduleQueryType.GET_SCHEDULES_BY_PROTAGONIST_ID:
        if query_params.protagonist_id is None:
            e = ValidationError(
                errors=[
                    ErrorWrapper(ValueError("field required"), "protagonist_id"),
                ],
                model=ScheduleGetQueryParams,
            )
            raise e
        stmt = (
            select(ScheduleDB)
            .where(ScheduleDB.protagonist_id == user_creds.user_id)
            .where(ScheduleDB.protagonist_id == query_params.protagonist_id)
        )
    if query_params.query_type == ScheduleQueryType.GET_SCHEDULES_BY_CLUSTER_ID:
        if query_params.cluster_id is None:
            e = ValidationError(
                errors=[
                    ErrorWrapper(ValueError("field required"), "protagonist_id"),
                ],
                model=ScheduleGetQueryParams,
            )
            raise e
        stmt = (
            select(ScheduleDB)
            .where(ScheduleDB.protagonist_id == user_creds.user_id)
            .where(ScheduleDB.assigned_cluster_id == query_params.cluster_id)
        )

    with DBSession() as session:
        try:
            schedule_db_list: List[ScheduleDB] = session.execute(stmt).scalars().all()
            schedule_list: List[Schedule] = [
                Schedule(schedule_id=obj.id, **obj.to_dict())
                for obj in schedule_db_list
            ]
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to fetch schedules due to {e}")
            raise e
    return schedule_list


def update_schedule(
    schedule_update_list: List[ScheduleUpdate], user_creds: User
) -> List[Schedule]:
    """Update schedule in the DB from ScheduleUpdate request.

    Args:
        schedule_update_list (List[ScheduleUpdate]): ScheduleUpdate request
        user_creds (User): User credentials.

    Raises:
        Exception: error

    Returns:
        List[ScheduleDB]: List of Instance of Updated Schedule
    """
    stmt = (
        select(ScheduleDB)
        .where(ScheduleDB.protagonist_id == user_creds.user_id)
        .where(
            ScheduleDB.id.in_(
                [
                    schedule_update.schedule_id
                    for schedule_update in schedule_update_list
                ]
            )
        )
    )
    with DBSession() as session:
        try:
            schedule_db_update_list: List[ScheduleDB] = (
                session.execute(stmt).scalars().all()
            )
            updated_schedule_list = []
            for schedule in schedule_db_update_list:
                for schedule_update in schedule_update_list:
                    if schedule_update.schedule_id == schedule.id:
                        schedule_update_dict = schedule_update.dict(
                            exclude_defaults=True
                        )
                        schedule_update_dict.pop("schedule_id")
                        updated_schedule = schedule.update(
                            session, **schedule_update_dict
                        )
                        updated_schedule_list.append(
                            Schedule(
                                schedule_id=updated_schedule.id,
                                **updated_schedule.to_dict(),
                            )
                        )
            return updated_schedule_list
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to update schedule: {schedule.to_dict()} due to {e}")
            raise e


def delete_schedule(
    schedule_delete_list: List[ScheduleDelete], user_creds: User
) -> List[ScheduleDB]:
    """Delete schedule in the DB from ScheduleDelete request.

    Args:
        schedule_delete_list (List[ScheduleDelete]): List of ScheduleDelete request
        user_creds (User): User credentials.

    Raises:
        Exception: error

    Returns:
        ScheduleDB: Instance of Deleted Schedule
    """
    stmt = (
        select(ScheduleDB)
        .where(ScheduleDB.protagonist_id == user_creds.user_id)
        .where(
            ScheduleDB.id.in_(
                [schedule.schedule_id for schedule in schedule_delete_list]
            )
        )
    )
    with DBSession() as session:
        try:
            schedule_db_delete_list: List[ScheduleDB] = (
                session.execute(stmt).scalars().all()
            )
            deleted_schedule_list = []
            for schedule in schedule_db_delete_list:
                schedule.delete(session)
                deleted_schedule_list.append(
                    Schedule(schedule_id=schedule.id, **schedule.to_dict())
                )
            return deleted_schedule_list
        except Exception as e:
            session.rollback()
            logger.error(f"Unable to delete schedule: {schedule.to_dict()} due to {e}")
            raise e


##################
# Flows
##################


def create_schedule_with_scheduler(inputs: List[Prop]) -> List[Prop]:
    """Create schedule db entry step function.

    Args:
        inputs (List[Prop]):
            Expects
                0: schedule_create_request_list
                    Prop(data=schedule_create_request_list, description="List of jobs to schedule") # noqa: B950
                2: user
                    Prop(data=user, description="User")

    Raises:
        FlowExecError: Flow execution failed

    Returns:
        List[Prop]:
            Created schedule objects

            Prop(data=created_schedule_list, description="List of created schedule Objects") # noqa: B950
    """
    try:
        schedule_create_request_list: List[ScheduleCreate] = inputs[0].data
        user: User = inputs[1].data
    except Exception as e:
        raise FlowExecError(
            human_description="Parsing inputs failed",
            error=e,
            error_type=type(e),
            is_user_facing=True,
            error_in_function=get_outer_function(),
        ) from e

    try:
        processed_schedule_create_request_list = [
            scheduler.schedule(schedule_create)
            for schedule_create in schedule_create_request_list
        ]
        created_schedule_list = create_schedule(
            processed_schedule_create_request_list, user
        )
    except SQLAlchemyError as e:
        raise FlowExecError(
            human_description="Creating DB object failed",
            error=e,
            error_type=type(e),
            is_user_facing=False,
            error_in_function=get_outer_function(),
        ) from e
    except Exception as e:
        raise FlowExecError(
            human_description="Something bad happened",
            error=e,
            error_type=type(e),
            is_user_facing=False,
            error_in_function=get_outer_function(),
        ) from e

    return [
        Prop(data=created_schedule_list, description="List of created schedule Objects")
    ]


def create_schedule_db_entry(inputs: List[Prop]) -> List[Prop]:
    """Create schedule db entry step function.

    Args:
        inputs (List[Prop]):
            Expects
                0: schedule_create_request_list
                    Prop(data=schedule_create_request_list, description="List of create schedule objects") # noqa: B950
                2: user
                    Prop(data=user, description="User")

    Raises:
        FlowExecError: Flow execution failed

    Returns:
        List[Prop]:
            Created schedule objects

            Prop(data=created_schedule_list, description="List of created schedule Objects") # noqa: B950
    """
    try:
        schedule_create_request_list: List[ScheduleCreate] = inputs[0].data
        user: User = inputs[1].data
    except Exception as e:
        raise FlowExecError(
            human_description="Parsing inputs failed",
            error=e,
            error_type=type(e),
            is_user_facing=True,
            error_in_function=get_outer_function(),
        ) from e

    try:
        created_schedule_list = create_schedule(schedule_create_request_list, user)
    except SQLAlchemyError as e:
        raise FlowExecError(
            human_description="Creating DB object failed",
            error=e,
            error_type=type(e),
            is_user_facing=False,
            error_in_function=get_outer_function(),
        ) from e
    except Exception as e:
        raise FlowExecError(
            human_description="Something bad happened",
            error=e,
            error_type=type(e),
            is_user_facing=False,
            error_in_function=get_outer_function(),
        ) from e

    return [
        Prop(data=created_schedule_list, description="List of created schedule Objects")
    ]


def revert_schedule_db_entry(inputs: List[Prop]) -> List[Prop]:
    """Delete db entry step function.

    Args:
        inputs (List[Prop]):
            Expects
                0: schedule_create_request_list
                    Prop(data=schedule_create_request_list, description="List of create schedule objects") # noqa: B950
                1: user
                    Prop(data=user, description="User")
                2: created_schedule_list
                    Prop(data=created_data_list, description="List of created schedule objects") # noqa: B950

    Raises:
        FlowExecError: Flow execution failed

    Returns:
        List[Prop]:
            Deleted schedule objects

            Prop(data=deleted_schedule_list, description="List of deleted schedule Objects") # noqa: B950
    """
    try:
        created_schedule_list: List[Schedule] = inputs[2].data
        user: User = inputs[1].data
    except Exception as e:
        raise FlowExecError(
            human_description="Parsing inputs failed",
            error=e,
            error_type=type(e),
            is_user_facing=True,
            error_in_function=get_outer_function(),
        ) from e

    try:
        deleted_schedule_list = delete_schedule(created_schedule_list, user)
    except SQLAlchemyError as e:
        raise FlowExecError(
            human_description="Deleting DB object failed",
            error=e,
            error_type=type(e),
            is_user_facing=False,
            error_in_function=get_outer_function(),
        ) from e
    except Exception as e:
        raise FlowExecError(
            human_description="Something bad happened",
            error=e,
            error_type=type(e),
            is_user_facing=False,
            error_in_function=get_outer_function(),
        ) from e

    return [
        Prop(data=deleted_schedule_list, description="List of deleted schedule Objects")
    ]


def publish_schedule_message(inputs: List[Prop]) -> List[Prop]:
    """Publish schedule message step function.

    Args:
        inputs (List[Prop]):
            Expects
                0: created_schedule_list
                    Prop(data=created_schedule_list, description="List of created schedule objects") # noqa: B950
                1: user
                    Prop(data=user, description="User")

    Raises:
        FlowExecError: Flow execution failed

    Returns:
        List[Prop]:
            Created schedule objects

            Prop(data=created_schedule_list, description="List of created schedule Objects") # noqa: B950
    """
    try:
        created_schedule_list: List[ScheduleCreate] = inputs[0].data
        user: User = inputs[1].data
    except Exception as e:
        raise FlowExecError(
            human_description="Parsing inputs failed",
            error=e,
            error_type=type(e),
            is_user_facing=True,
            error_in_function=get_outer_function(),
        ) from e

    try:
        for schedule in created_schedule_list:
            messenger = messenger_registry.get_messenger(schedule.messenger)
            job = get_job(
                JobGetQueryParams(
                    query_type=JobQueryType.GET_JOBS_BY_ID, job_id=schedule.job_id
                ),
                user,
            )
            message = Message(data=job, msg_type=MessageType.TO_DEMON_SCHEDULE_MSG)
            messenger.publish(message, schedule.messenger_queue)
    except Exception as e:
        raise FlowExecError(
            human_description="Something bad happened",
            error=e,
            error_type=type(e),
            is_user_facing=False,
            error_in_function=get_outer_function(),
        ) from e

    return [
        Prop(data=created_schedule_list, description="List of created schedule Objects")
    ]


def generate_create_schedule_flow(
    schedule_create_request_list: List[Job], user: User
) -> Flow:
    """Generate flow for creating the schedule resource.

    Args:
        schedule_create_request_list (List[ScheduleCreate]): List of schedule create objects.
        user (User): User credentials.

    Returns:
        Flow: Create schedule flow
    """
    create_schedule_with_scheduler_step = Step(
        create_schedule_with_scheduler, dummy_backward
    )
    create_db_entry_step = Step(create_schedule_db_entry, revert_schedule_db_entry)
    publish_schedule_message_step = Step(publish_schedule_message, dummy_backward)

    flow = Flow()

    flow.add_step(
        create_schedule_with_scheduler_step,
        [
            Prop(
                data=schedule_create_request_list,
                description="List of schedule create request objects",
            ),
            Prop(data=user, description="User"),
        ],
    )

    create_db_entry_step_func_prop = FuncProp(
        func=flow.get_froward_output,
        params={"index": 0},
        description="Created schedule create object list",
    )
    flow.add_step(
        create_db_entry_step,
        [
            create_db_entry_step_func_prop,
            Prop(data=user, description="User"),
        ],
    )

    publish_schedule_message_step_func_prop = FuncProp(
        func=flow.get_froward_output,
        params={"index": 1},
        description="Created schedule list",
    )
    flow.add_step(
        publish_schedule_message_step,
        [
            publish_schedule_message_step_func_prop,
            Prop(data=user, description="User"),
        ],
    )

    return flow



"""Blackcap Auth PUT route schemas."""

from typing import Any, Dict, List, Union

from blackcap.schemas.api.common import ResponseSchema
from blackcap.schemas.user import User
from pydantic import BaseModel


class AuthCredsUpdate(BaseModel):
    """Schema to parse update user creds requests."""

    update_token: str
    new_password: str


class AuthPOSTResponse(ResponseSchema):
    """User POST response schema."""

    items: Dict[str, List[Union[User, Any]]]



"""NATS implementation of messenger."""

from dataclasses import asdict
import json
from typing import Callable, Dict, Optional

from logzero import logger
from pynats import NATSClient, NATSMessage

from blackcap.configs.base import BaseConfig
from blackcap.messenger.base import BaseMessenger
from blackcap.schemas.message import Message
from blackcap.utils.json_encoders import UUIDEncoder


class NATSMessenger(BaseMessenger):
    """NATS implementation of Messenger."""

    CONFIG_KEY_VAL = "NATS"

    def __init__(self: "NATSMessenger", config: BaseConfig) -> None:
        """Initialize Messenger with app config.

        Args:
            config (BaseConfig): Config to initialize messenger
        """
        self.config = config

    @property
    def client(self: "NATSMessenger") -> NATSClient:
        """NATS client object."""
        client = NATSClient(url=self.config.NATS_ENDPOINT, name=self.config.FLASK_APP)
        client.connect()
        return client

    def publish(self: "NATSMessenger", msg: Dict, topic_id: str) -> str:
        """Publish msg on the GCP Pub/Sub queue.

        Args:
            msg (Dict): Messsag to publish
            topic_id (str): Id of the topic

        Returns:
            str: Id of the published msg
        """
        # Msg must be a bytestring
        msg = json.dumps(msg, cls=UUIDEncoder).encode("utf-8")
        self.client.publish(subject=topic_id, payload=msg)
        self.client.close()
        return "ok"

    def subscribe(
        self: "NATSMessenger",
        callback: Callable,
        sub_id: str,
        timeout: Optional[float] = None,
    ) -> None:
        """Subscribe to a topic.

        Args:
            callback (Callable): Callback to invoke when a msg is received
            sub_id (str): Id of the topic.
            timeout (Union[float, None]): Time to wait for msgs. Defaults to None. # noqa: E501
        """
        try:
            client = self.client
            sub = client.subscribe(subject=sub_id, callback=callback)
            logger.info(f"Subscription created: {sub}")
            client.wait(count=None)
            client.close()
        except Exception as e:
            logger.error(
                f"NATSMessenger subscribe error while pulling messages. Error: {e}"
            )
            client.close()

    def parse_messenger_msg(
        self: "NATSMessenger", messenger_msg: NATSMessage
    ) -> Message:
        """Parse messenger msg to blackcap mesage schema.

        Args:
            messenger_msg (NATSMessage): NATSMessenger message

        Returns:
            Message: Parsed Message
        """
        return Message.parse_obj(asdict(messenger_msg))

    def echo_msg(self: "NATSMessenger", msg: NATSMessage) -> None:
        """Echo msgs to stdout.

        Args:
            msg (NATSMessage): Message to echo
        """
        print(asdict(msg))



"""Blackcap job GET route."""

from http import HTTPStatus

from flask import make_response, request, Response
from pydantic.error_wrappers import ValidationError
from sqlalchemy.exc import SQLAlchemyError

from blackcap.blocs.cluster import get_cluster
from blackcap.routes.cluster import cluster_bp
from blackcap.schemas.api.cluster.get import ClusterGetQueryParams, ClusterGetResponse


@cluster_bp.get("/")
def get() -> Response:
    """Get cluster.

    Returns:
        Response: Flask response
    """
    # Parse query params from request
    try:
        query_params = ClusterGetQueryParams.parse_obj(request.args)
    except ValidationError as e:
        response_body = ClusterGetResponse(
            msg="query validation error", errors={"main": e.errors()}
        )
        return make_response(response_body.json(), HTTPStatus.BAD_REQUEST)
    except Exception:
        response_body = ClusterGetResponse(
            msg="unknown error", errors={"main": ["unknown internal error"]}
        )
        return make_response(response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR)

    # Get cluster list from the DB
    try:
        cluster_list = get_cluster(query_params)
    except SQLAlchemyError:
        response_body = ClusterGetResponse(
            msg="internal databse error", errors={"main": ["unknown internal error"]}
        )
        return make_response(response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR)
    except Exception:
        response_body = ClusterGetResponse(
            msg="unknown error", errors={"main": ["unknown internal error"]}
        )
        return make_response(response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR)

    # return fetched cluster list in response
    response_body = ClusterGetResponse(
        msg="cluster list successfully retrieved", items={"cluster_list": cluster_list}
    )
    return make_response(response_body.json(), HTTPStatus.OK)



"""CLI commnad util functions."""


import subprocess  # noqa: S404
import sys
from typing import List, Optional, Union

from blackcap.configs import config_registry
from blackcap.configs.base import BaseConfig

config = config_registry.get_config()


def call_cli(
    cmd: List[str],
    input: Optional[str] = None,
    timeout: Optional[float] = None,
    config: BaseConfig = config,
) -> Union[str, bytes]:
    """Call CLI commands.

    Args:
        cmd (List[str]): [description]
        input (Optional[str]): [description]. Defaults to None.
        timeout (Optional[float]): [description]. Defaults to None.
        config (BaseConfig): [description]. Defaults to default_config.

    Returns:
        Union[str, bytes]: Stdout
    """
    proc = subprocess.Popen(  # noqa: S603
        cmd,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        encoding="utf-8",
        errors="replace",
    )
    stdout, stderr = proc.communicate(input=input, timeout=timeout)
    sys.stderr.write(stderr)
    return stdout



"""Blackcap User PUT route schemas."""

from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel
from pydantic.types import UUID4

from blackcap.schemas.api.common import ResponseSchema
from blackcap.schemas.user import User


class UserUpdate(BaseModel):
    """Schema to parse update user requests."""

    user_id: UUID4
    name: Optional[str]
    email: Optional[str]
    organisation: Optional[str]


class UserPUTResponse(ResponseSchema):
    """User PUT response schema."""

    items: Dict[str, List[Union[User, Any]]] = {}



"""Flow Object."""

from dataclasses import dataclass, field
from enum import auto, Enum, unique
import inspect
from typing import Any, List, Optional, Union

from blackcap.flow.step import FuncProp, Prop, Step


class AutoName(Enum):
    """Helper class to auto name enum."""

    def _generate_next_value_(
        name: Any, start: Any, count: Any, last_values: Any  # noqa: B902
    ) -> str:
        """Generate names."""
        return name


@unique
class FlowStatus(AutoName):
    """Status of the Flow."""

    CREATED = auto()
    EXECUTING = auto()
    PASSED = auto()
    FAILED = auto()


@dataclass
class FlowExecError(Exception):
    """Errors while executing flow."""

    human_description: str
    error: Any
    # TODO: Change it to enum?
    error_type: type
    error_in_function: str
    step_index: Optional[int] = None
    is_user_facing: bool = False
    user_facing_msg: Optional[str] = None


class FlowBuildError(Exception):
    """Errors while building flow."""

    human_description: str
    error: Any
    # TODO: Change it to enum?
    error_type: type
    error_in_function: str


def get_outer_function() -> str:
    """Return name of the outer function."""
    return inspect.getframeinfo(inspect.currentframe().f_back).function


@dataclass
class Flow:
    """Flow Object."""

    steps: List[Step] = field(default_factory=list)
    inputs: List[List[Union[Prop, FuncProp]]] = field(default_factory=list)
    forward_outputs: List[List[Prop]] = field(default_factory=list)
    backward_outputs: List[List[Prop]] = field(default_factory=list)
    status: FlowStatus = FlowStatus.CREATED
    errors: List[FlowExecError] = field(default_factory=list)

    def add_step(self: "Flow", step: Step, inputs: List[Prop]) -> None:
        """Add a step to the flow."""
        # TODO: Add checks in future
        self.steps.append(step)
        self.inputs.append(inputs)

    def get_froward_output(self: "Flow", index: int) -> Optional[List[Prop]]:
        """Lazily get list of prop from flow forward outputs."""
        return self.forward_outputs[index]

    def get_input(self: "Flow", index: int) -> Optional[List[Prop]]:
        """Lazily get list of prop from flow inputs."""
        return self.inputs[index]



"""Blackcap Job PUT route schemas."""

from typing import Any, Dict, List, Union

from pydantic import BaseModel
from pydantic.types import UUID4

from blackcap.schemas.job import Job


class JobUpdate(BaseModel):
    """Schema to parse update job requests."""

    job_id: UUID4


class JobPUTRequest(BaseModel):
    """Job PUT request schema."""

    job_list: List[JobUpdate]


class JobPUTResponse(BaseModel):
    """Job PUT response schema."""

    items: Dict[str, List[Union[Job, Any]]] = {}



"""Blackcap job GET route."""

from http import HTTPStatus

from flask import make_response, request, Response
from pydantic.error_wrappers import ValidationError
from sqlalchemy.exc import SQLAlchemyError

from blackcap.blocs.job import get_job
from blackcap.routes.job import job_bp
from blackcap.schemas.api.job.get import JobGetQueryParams, JobGetResponse
from blackcap.schemas.user import User
from blackcap.utils.auth import check_authentication


@job_bp.get("/")
@check_authentication
def get(user: User) -> Response:
    """Get job.

    Args:
        user (User): Extracted user from request

    Returns:
        Response: Flask response
    """
    # Parse query params from request
    try:
        query_params = JobGetQueryParams.parse_obj(request.args)
    except ValidationError as e:
        response_body = JobGetResponse(
            msg="query validation error", errors={"main": e.errors()}
        )
        return make_response(response_body.json(), HTTPStatus.BAD_REQUEST)
    except Exception:
        response_body = JobGetResponse(
            msg="unknown error", errors={"main": ["unknown internal error"]}
        )
        return make_response(response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR)

    # Get jobs from the DB
    try:
        job_list = get_job(query_params, user)
    except SQLAlchemyError:
        response_body = JobGetResponse(
            msg="internal databse error", errors={"main": ["unknown internal error"]}
        )
        return make_response(response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR)
    except Exception:
        response_body = JobGetResponse(
            msg="unknown error", errors={"main": ["unknown internal error"]}
        )
        return make_response(response_body.json(), HTTPStatus.INTERNAL_SERVER_ERROR)

    # return fetched jobs in response
    response_body = JobGetResponse(
        msg="job successfully retrieved", items={"job_list": job_list}
    )
    return make_response(response_body.json(), HTTPStatus.OK)



"""Slurm interface."""

from os import makedirs
from pathlib import Path
from typing import List

from xdg import xdg_data_home

from blackcap.cluster.base import BaseCluster
from blackcap.schemas.schedule import Schedule
from blackcap.utils.cli_commands import call_cli


class SlurmCluster(BaseCluster):
    """Slurm interface."""

    CONFIG_KEY_VAL = "SLURM"

    def prepare_job(self: "SlurmCluster", schedule: Schedule) -> None:
        """Prepare job for submission.

        Args:
            schedule (Schedule): Schedule Object
        """
        job = schedule.job
        job_data_path: Path = (
            xdg_data_home() / "orchestra" / "demon" / "jobs" / str(job.job_id)
        )
        makedirs((job_data_path / "out"), exist_ok=True)
        with open(job_data_path.joinpath("start.sh"), "w+") as f:
            f.write(job.script)
        with open(job_data_path.joinpath("notify_start.sh"), "w+") as f:
            f.write(
                f"""#! /bin/bash
                conda activate orchestra

                demon pub schedup --sched_id={schedule.schedule_id} --job_id={job.job_id} --status=RUNNING  # noqa: B950
                """
            )
        with open(job_data_path.joinpath("notify_ok.sh"), "w+") as f:
            f.write(
                f"""#! /bin/bash
                conda activate orchestra

                demon pub schedup --sched_id={schedule.schedule_id} --job_id={job.job_id} --status=COMPLETED  # noqa: B950
                """
            )
        with open(job_data_path.joinpath("notify_not_ok.sh"), "w+") as f:
            f.write(
                f"""#! /bin/bash
                conda activate orchestra

                demon pub schedup --sched_id={schedule.schedule_id} --job_id={job.job_id} --status=FAILED  # noqa: B950
                """
            )

    def submit_job(self: "SlurmCluster", schedule: Schedule) -> str:
        """Submit job to the cluster.

        Args:
            schedule (Schedule): Schedule Object

        Returns:
            str: Job ID
        """
        job = schedule.job
        self.prepare_job(schedule)
        job_data_path: Path = (
            xdg_data_home() / "orchestra" / "demon" / "jobs" / str(job.job_id)
        )
        job_script_path: Path = job_data_path / "start.sh"
        # notify_start_script_path: Path = job_data_path / "notify_start.sh"
        # notify_ok_script_path: Path = job_data_path / "notify_ok.sh"
        # notify_not_ok_script_path: Path = job_data_path / "notify_not_ok.sh"
        job_out_path: Path = job_data_path / "out"
        main_job_cmd_args_list = [
            "sbatch",
            job_script_path.absolute(),
            "-o",
            job_out_path.absolute(),
        ]
        output = call_cli(main_job_cmd_args_list)

        # example: Submitted batch job 45
        main_job_id = output.split(" ")[-1]
        # TODO: Fix notify scripts
        # Add notify jobs
        # notify_start_job_cmd_args_list = [
        #     "sbatch",
        #     f"--dependency=after:{main_job_id}",
        #     notify_start_script_path.absolute(),
        #     "-o",
        #     job_out_path.absolute(),
        # ]
        # call_cli(notify_start_job_cmd_args_list)

        # notify_ok_job_cmd_args_list = [
        #     "sbatch",
        #     f"--dependency=afterok:{main_job_id}",
        #     notify_ok_script_path.absolute(),
        #     "-o",
        #     job_out_path.absolute(),
        # ]
        # call_cli(notify_ok_job_cmd_args_list)

        # notify_not_ok_job_cmd_args_list = [
        #     "sbatch",
        #     f"--dependency=afternotok:{main_job_id}",
        #     notify_not_ok_script_path.absolute(),
        #     "-o",
        #     job_out_path.absolute(),
        # ]
        # call_cli(notify_not_ok_job_cmd_args_list)
        return main_job_id

    def get_job_status(self: "SlurmCluster", job_id: str) -> List[str]:
        """Get status of a job by Job.

        Args:
            job_id (str): ID of the job

        Returns:
            List[str]: List of status of the jobs
        """
        cmd_args_list = [
            "sacct",
            "-S",
            "1970-01-02",
            "--format",
            "State,ExitCode",
            "-j",
            job_id,
        ]
        output = call_cli(cmd_args_list)
        job_status_list = [status for status in output.strip().split("\n")[2:]]
        return job_status_list
